open Interval
open Box
(* open Constr *)

type variable = int      (* should be treated as hidden, and only known to prune.ml *)
type predicateTernary = Addition | Multiplication | Subtraction;;
type predicateBinary = Sin | Cos | Tan | ArcSin | ArcCos | ArcTan | Exp | Log | Square | UMinus;;
type predicateUnary = Equal | LessEqual | Less | NotEqual | Greater |
GreaterEqual ;;

type primTernary = { predicate3: predicateTernary; var31: variable; var32:variable; var33 : variable };;
type primBinary = { predicate2: predicateBinary; var21: variable; var22: variable; };;
type primUnary = { predicate1: predicateUnary; var1: variable };;
type primPower = { var01:variable; n:int ;var02:variable;};;
type isElement = { var: variable; interval: interval;};;
type posSemDef = variable array array 

(* variable numbering: first free variables, number of introduction when going from root to leaves; so quantifiers in different paths share variable numbers *)

type constr  =   
    BoolVar of int 
  | UserDefined of (box -> unit) * (int -> bool)
  | BoolIntv of Bool_intv.bintv 
  | Ternary of primTernary 
  | Binary of primBinary 
  | Unary of primUnary 
  | Power of primPower 
  | IsElement of isElement 
  | SubSet of subset
  | PosSemDef of posSemDef 
  | Atomic of atomic 
  | Conjunction of connective 
  | Disjunction of connective 
  | UnivQuant of quantification 
  | ExistQuant of quantification
  | Redundant of redundant 
  | UnivQuantBound of quantbound

and quantification = 
    { qbound: box; 
      subconstr: constr; 
      mutable outer_changed: int list; (* use bit-vector instead? *)
      mutable inner_changed: int list; (* use bit-vector instead? *)
      mutable qnext_branch: int } 
and connective =
    { mutable branches: constr list; 
      mutable next_branch: int; 
      mutable bloated: bool; 
      mutable changed: int list; 
      mutable wakeup_table: int list array;
      mutable maxconj_pos: bool;
      mutable maxconj_neg: bool
    }
and atomic =
    { obound: box; ibound: box; aconstr: constr } 
and subset = 
    { subset_var1: variable; subset_var2: variable }
and redundant =
    { original_constr: constr; redundant_constr: constr }
and quantbound =
    { var_trans: variable array; qbsubconstr: constr }


type bconstr = constr * box

(*
let rec max_var (constr: constr): int =
  match constr with
    BoolVar _ -> -1 |
    BoolIntv _ -> -1 |
    Unary c -> c.var1 |
    Binary c -> max c.var21 c.var22 |
    Ternary c -> max c.var31 (max c.var32 c.var33) |
    Atomic a -> max_var a.aconstr | 
    Power c -> max c.var01 c.var02 |
    SubSet c -> max c.subset_var1 c.subset_var2 |
    IsElement c -> c.var |
    PosSemDef p -> Array.fold_left max 0 (Array.map (function a -> Array.fold_left max 0 a) p) |
    Conjunction c -> List.fold_left max 0 (List.map max_var c.branches) |
    Disjunction c -> List.fold_left max 0 (List.map max_var c.branches) |
    UnivQuant c -> max_var c.subconstr |
    ExistQuant c -> max_var c.subconstr |
    UnivQuantBound c -> max_var c.qbsubconstr 
  | Redundant r -> max (max_var r.original_constr) (max_var r.redundant_constr)
*)
(*    _ -> assert(false) *)

let linear_relaxation= ref false

let rec max_width (constr: constr) =
  match constr with 
    BoolVar _ | UserDefined _  | BoolIntv _ | Unary _ | Binary _ | Ternary _ | Power _ | IsElement _ | SubSet _ | PosSemDef _ -> 0.0 
  | Atomic a -> max_width a.aconstr |
    Conjunction c -> List.fold_left max 0. (List.map max_width c.branches) |
    Disjunction c -> List.fold_left max 0. (List.map max_width c.branches) |
    UnivQuant c -> 
      max (width c.qbound) (max_width c.subconstr) |
    ExistQuant c -> 
      max (width c.qbound) (max_width c.subconstr) 
  | Redundant c ->
      max (max_width c.original_constr) (max_width c.redundant_constr)
  | UnivQuantBound c ->
      max_width c.qbsubconstr

(*let complete_box (b: box) (c: constr)=
      cart_prod2 b (full_box (((max_var c) - (dim b))+1)) *)

let neg_unary (p: predicateUnary) =
  match p with
    Equal -> NotEqual |
    LessEqual -> Greater |
    Less -> GreaterEqual |
    NotEqual -> Equal |
    Greater -> LessEqual |
    GreaterEqual -> Less

let neg_atomic (c: constr): constr =
  let neg_pred (c: constr): constr =
    match c with
      Unary c  -> Unary { c with predicate1= neg_unary c.predicate1 }
    | c -> c in
  (* begin neg_atomic *)
  match c with 
    Conjunction c -> Conjunction { c with branches= List.map neg_pred c.branches }
  | _ -> assert(false)

let neg (c: constr): constr =
  match c with 
    Atomic a -> Atomic { a with aconstr= neg_atomic a.aconstr }
  | Conjunction c -> Disjunction c
  | Disjunction c -> Conjunction c
  | UnivQuant c -> ExistQuant c
  | ExistQuant c -> UnivQuant c
  | _ -> assert(false)

let rec copy (c: constr) = 
  (* the mutable fields of the 'constr' type have to be copied explicitely! *)
  match c with 
    ExistQuant { qbound= qb; subconstr= sc; outer_changed= ocons; inner_changed= icons; qnext_branch= n } -> 
      ExistQuant { qbound= Box.copy qb; subconstr= copy sc; outer_changed= ocons; inner_changed= icons; qnext_branch= n } |
    UnivQuant { qbound= qb; subconstr= sc; outer_changed= ocons; inner_changed= icons; qnext_branch= n } ->  
      UnivQuant { qbound= Box.copy qb; subconstr= copy sc; outer_changed= ocons; inner_changed= icons; qnext_branch= n } | 
    Atomic { obound= ob; ibound= ib; aconstr= ac } -> 
      Atomic { obound= Box.copy ob; ibound= Box.copy ib; aconstr= copy ac } |
    Disjunction scl -> 
      Disjunction { branches= (List.map copy scl.branches); bloated= scl.bloated; next_branch= scl.next_branch; changed= scl.changed; wakeup_table= scl.wakeup_table; maxconj_pos= scl.maxconj_pos; maxconj_neg= scl.maxconj_neg }
  |
    Conjunction scl -> 
      Conjunction { branches= (List.map copy scl.branches); bloated= scl.bloated; next_branch= scl.next_branch; changed= scl.changed; wakeup_table= scl.wakeup_table; maxconj_pos= scl.maxconj_pos; maxconj_neg= scl.maxconj_neg } 
  | Redundant c -> 
      Redundant { original_constr= copy c.original_constr; redundant_constr= copy c.redundant_constr }
  | UnivQuantBound q -> UnivQuantBound { q with qbsubconstr= copy q.qbsubconstr }
  | (BoolVar _ | UserDefined _ | BoolIntv _ | IsElement _ | Power _ | Unary _ | Binary _ | Ternary _ | SubSet _ | PosSemDef _ as c) -> c

let max_width_bounded ((c, b): bconstr) =
  max (width b) (max_width c)

let max_width_q (q: quantification) =
  max (width q.qbound) (max_width q.subconstr)

let rec aux_consistent (c: constr): bool =
(* checks whether bounds of auxiliary variables are consistent after full pruning *)
  match c with
    Conjunction l -> List.for_all (function cp -> aux_consistent cp) l.branches |
    Disjunction l -> List.for_all (function cp -> aux_consistent cp) l.branches |
    ExistQuant q -> is_empty q.qbound ||  (aux_consistent q.subconstr) |
    UnivQuant q -> is_empty q.qbound ||  (aux_consistent q.subconstr) |
    Atomic a -> List.for_all (function k -> not (Interval.is_empty (intersect (get_bound a.obound k) (get_bound a.ibound k)))) (Util.int_list 0 ((dim a.obound)-1)) |
    _ -> true

let rec is_consistent_free (c: constr) (free: int) =
  match c with
    Unary c -> 
(*	print_endline "*unary*"; print_int c.var1; print_string " "; print_int free; *)
      c.var1<free 
  | Binary c -> 
      (c.var21<free && c.var22<free) 
  | Ternary c -> 
      (c.var31<free && c.var32<free && c.var33<free) 
  | Conjunction l -> 
	(Array.length l.wakeup_table)<=free &&
	let is_constr v = v<List.length l.branches in
	List.for_all is_constr l.changed &&
	Array.fold_left (&&) true (Array.map (function l -> List.for_all is_constr l) l.wakeup_table) &&
	List.for_all (function cp -> is_consistent_free cp free) l.branches 
  | Disjunction l -> 
	(Array.length l.wakeup_table)<=free &&
	let is_constr v = v<List.length l.branches in
	List.for_all is_constr l.changed &&
	Array.fold_left (&&) true (Array.map (function l -> List.for_all is_constr l) l.wakeup_table) &&
	List.for_all (function cp -> is_consistent_free cp free) l.branches |
	Atomic a -> (dim a.obound=dim a.ibound) && 
          is_consistent_free a.aconstr (free+(dim a.obound)) 
  | ExistQuant q -> 
	is_empty q.qbound || 
                      (is_consistent_free q.subconstr (free+(dim q.qbound)) && 
		       Util.unique q.outer_changed && List.for_all (function v-> v>=free) q.outer_changed && 
		       Util.unique q.inner_changed && List.for_all (function v-> v>=free) q.inner_changed) 
  | UnivQuant q -> 
	is_empty q.qbound || (is_consistent_free q.subconstr (free+(dim q.qbound)) && Util.unique q.outer_changed && Util.unique q.inner_changed) |
      _ -> true
  


let is_consistent (c: constr) (b: box) =
  (is_empty b) || ((Box.is_consistent b) && (is_consistent_free c (dim b)))


         let rec print_space i=
               if i = 0 then
                ()
               else
                 begin
                   print_string " ";print_space (i-1)
                 end

         let print_power c =
                print_string "Power {var01 = ";print_int c.var01;
                print_string "; n = ";print_int c.n;
                print_string "; var02 = ";print_int c.var02;
                print_string "; }\n"

         let print_pred_unary t=
	        match t with
		Equal->print_string "Equal"; |
                LessEqual -> print_string "LessEqual"; |
		Less -> print_string "Less"; |
		NotEqual->print_string "NotEqual"; |
		Greater->print_string "Greater"; |
		GreaterEqual->print_string "GreaterEqual"

	 let print_pred_binary t=
	        match t with
		Sin->print_string "Sin"; |
                Cos -> print_string "Cos"; |
                Tan -> print_string "Tan"; |
                ArcSin -> print_string "ArcSin"; |
                ArcCos -> print_string "ArcCos"; |
                ArcTan -> print_string "ArcTan"; |
		Exp -> print_string "Exp"; |
                Log -> print_string "Log"; |
		Square->print_string "Square"; 
	      | UMinus->print_string "UMinus"
		
	 let print_pred_ternary t=
	        match t with
		Addition->print_string "Addition"; |
                Multiplication -> print_string "Multiplication"; |
		Subtraction -> print_string "Subtraction"

	 let print_iselement c =
                print_string "IsElement {var= ";print_int c.var;
                print_string "; interval= "; Interval.print c.interval;
	        print_string "; }\n"

	 let print_unary c =
	        print_string "Unary {predicate1= ";print_pred_unary c.predicate1;
                print_string "; var1= ";print_int c.var1;
	        print_string "; }\n"

         let print_binary c =
	        print_string "Binary {predicate2= ";print_pred_binary c.predicate2;
                print_string "; var21= ";print_int c.var21;
                print_string "; var22= ";print_int c.var22;
	        print_string "; }\n"

         let print_ternary c =
	        print_string "Ternary {predicate3= ";print_pred_ternary c.predicate3;
		print_string "; var31= ";print_int c.var31;
          	print_string "; var32= ";print_int c.var32;
          	print_string "; var33= ";print_int c.var33;
	  	print_string "; }\n"
 
         let print_subset c =
	        print_string "SubSet { var21= ";print_int c.subset_var1;
                print_string "; var22= ";print_int c.subset_var2;
	        print_string "; }\n"


	 let print_possemdef (p: posSemDef) =
	   print_string "PosSemDef "; 
	   for i= 0 to (Array.length p)-1 do
	     for j= 0 to (Array.length p.(i))-1 do
	       print_int p.(i).(j);
	       if j<(Array.length p)-1 then print_char ' ';
	     done;
	     if i<(Array.length p)-1 then print_string ", ";
	   done;
	   print_endline ""

         (*let print_bconstr c =*)
                  
         let print_quant indent c v d forall=      
                if forall then begin
                  print_space indent;
                  print_string "FORALL";
                  print_string " :";
                   for i = v to v + d-1 do 
                   print_int i;print_string "  ";
                  done;
                  print_string " ";
                  Box.print c.qbound;
		  print_endline "";
                end else begin
                    print_space indent;
                    print_string "EXISTS";
                    print_string " : ";
                    for i =v to v + d-1 do
                      print_int i;print_string "  ";
                    done;
                    print_string " ";
                    Box.print c.qbound;
		    print_int c.qnext_branch;
		    print_endline "";
                  end

         let print_atomic indent c v d =      
           begin
             print_space indent;
             print_string "ATOMIC";
             print_string " : ";
             for i =v to v + d-1 do
               print_int i;print_string "  ";
             done;
             print_string " ";
             Box.print c.obound;
             print_string " ";
             Box.print c.ibound;
	     print_endline "";
           end

                        
                 
let rec print_constraint c indent v=
  let print_wakeup_table t =
    for i= 0 to (Array.length t)-1 do
      print_char '[';
      List.iter (function i -> print_int i; print_string " ") t.(i);
      print_char ']';
    done in
  match c with
    BoolVar k -> print_space (indent+2); print_endline "BoolVar"; print_int k |
    BoolIntv iv -> print_space (indent+2); print_string "BoolIntv:"; Bool_intv.print iv; print_newline () |
    UserDefined _ -> print_space (indent+2); print_string "UserDefined:"; print_newline ()
  | Ternary c->print_space (indent+2);print_ternary c; |
    Binary c->print_space (indent+2);print_binary c; |
    Unary c->print_space (indent+2); print_unary c; |
    Power c->print_space (indent+2); print_power c; |
    IsElement c->print_space (indent+2); print_iselement c; |
    SubSet c -> print_space (indent+2);print_subset c; |
    PosSemDef p -> print_space (indent+2); print_possemdef p; |
    Atomic a -> print_atomic indent a v (Box.dim a.obound); print_constraint a.aconstr (indent+2) ((Box.dim a.obound) + v );| 
    ExistQuant c->print_quant indent c v (Box.dim c.qbound) false ;print_constraint c.subconstr (indent+2) ((Box.dim c.qbound) + v );| 
    UnivQuant c->print_quant indent c v (Box.dim c.qbound) true;print_constraint c.subconstr (indent+2) ((Box.dim c.qbound) + v );  |
    Conjunction c->begin
      print_space indent;
      print_string "Conjunction "; print_wakeup_table c.wakeup_table; 
      if c.maxconj_pos then print_string "+";
      if c.maxconj_neg then print_string "-";
      print_newline ();
      print_space (indent+1);print_string "[\n";
      print_logicop c (indent+2) v;
      print_space (indent+2); print_string "]\n"
    end     |
    
    Disjunction c->begin
      print_space indent;
      print_string "Disjunction "; print_int (Array.length c.wakeup_table); print_newline ();
      print_space (indent+1);print_string "[\n";
      print_logicop c (indent+2) v;
      print_space (indent+2); print_string "]\n"
    end
  | Redundant c -> 
      print_constraint c.original_constr indent v;
      let print_redundant c = begin
	print_space indent; print_endline "Redundant: ";
	print_constraint c (indent+2) v
      end in
      print_redundant c.redundant_constr
  | UnivQuantBound c ->
      print_space indent; print_string "UnivQuantBound "; 
      let print_trans v = print_string ","; print_int v; in
      Array.iter print_trans c.var_trans; print_newline ();
      print_constraint c.qbsubconstr (indent+2) (v+Array.length c.var_trans)
	
	
and print_logicop c indent v=
  match c.branches with
    []-> ()
  |head::tail->begin
      print_constraint head indent v;
      print_logicop { c with branches=tail } indent v
  end 
        
	
	
	
let print_constr (c: constr) =
  print_constraint c 0 0

let print_bconstr ((c, b): bconstr) = 
  begin
    Box.print b;
    print_constraint c 0 (dim b); 
  end

let rec contains_var (c: constr) (v: variable): bool =
  let cv constr= contains_var constr v in
  match c with
    BoolVar _ -> false |
    BoolIntv _ -> false |
    Unary c -> v==c.var1 |
    Binary c -> (v== c.var21) || (v==c.var22) |
    Ternary c -> (v==c.var31) || (v==c.var32) || (v==c.var33) |
    Power c -> (v==c.var01) || (v==c.var02) |
    IsElement c -> (v==c.var) |
    SubSet c -> (v==c.subset_var1) || (v==c.subset_var2) |
    PosSemDef p -> Array.fold_left (||) false (Array.map (function a -> Array.fold_left (||) false (Array.map (function av -> av=v) a)) p) |
    Conjunction c -> List.exists cv c.branches |
    Disjunction c -> List.exists cv c.branches |
    UnivQuant c -> cv c.subconstr |
    ExistQuant c -> cv c.subconstr |
    Atomic a -> cv a.aconstr 
  | Redundant c -> cv c.original_constr || cv c.redundant_constr
  | UnivQuantBound c -> cv c.qbsubconstr
  | UserDefined (_, udf_var) -> udf_var v
      


(*
let rec consistent_marks (c: constr) (free_vars: int) (outer: bool) (cons: int list): bool=
  let cons_branch c = consistent_marks c free_vars outer cons
  and cons_var (v: variable):bool = (not (contains_var c v)) || (List.mem v cons) in
  List.for_all cons_var (Util.int_list 0 (free_vars-1)) &&
    match c with 
      Ternary _ -> true |
      Binary _ -> true | 
      Unary _ -> true | 
      IsElement _ -> true | 
      Power _ -> true | 
      Conjunction c -> List.for_all cons_branch c.branches | 
      Disjunction c -> List.for_all cons_branch c.branches |
      UnivQuant q -> consistent_marks q.subconstr (free_vars+(dim q.qbound)) outer ((if outer then q.outer_changed else q.inner_changed) @ cons) |
      ExistQuant q -> 
	if q.qnext_branch>=0 then
	  consistent_marks q.subconstr (free_vars+(dim q.qbound)) outer ((if outer then q.outer_changed else q.inner_changed) @ cons)
	else (* atomic constraints are consistent if free variables unchanged, this might change when we keep intervals for artificial variables *)
	  true
*)

let make_wakeup_table (t: constr list) (n: int): int list array = 
  Array.init n (function v -> Util.find_index (function c-> contains_var c v) t) 

let connective_remove (conn: connective) (i: int): unit =
  let rec index_remove (i: int) (l: int list): int list =
    match l with
      [] -> [] |
      k::r -> 
	let rp = index_remove i r in
	if k<i then k::rp 
	else if k=i then rp
	else (k-1)::rp in
  conn.branches<- Util.drop_nth conn.branches i;
  conn.changed<- index_remove i conn.changed;
  conn.wakeup_table<- Array.map (index_remove i) conn.wakeup_table

let rec clear (c: constr): unit =
  match c with 
    BoolVar _
  | UserDefined _
  | BoolIntv _
  | Unary _ 
  | Binary _  
  | Ternary _
  | Power _
  | SubSet _
  | IsElement _
  | PosSemDef _ -> () 
  | ExistQuant q -> 
(*      assert(false); *)
      Box.set_full q.qbound;
      clear q.subconstr
  | UnivQuant q ->  
(*      assert(false);*)
      Box.set_full q.qbound;
      clear q.subconstr
  |  Atomic a  -> begin
      Box.set_full a.obound; 
      Box.set_full a.ibound;
    end 
  | Disjunction scl -> begin
      List.iter clear scl.branches;
      scl.bloated <- true;
      scl.next_branch<- 0;
      scl.changed<- Util.int_list 0 ((List.length scl.branches)-1); 
    end 
  | Conjunction scl -> begin
      List.iter clear scl.branches;
      scl.bloated <- true;
      scl.next_branch<- 0;
      scl.changed<- Util.int_list 0 ((List.length scl.branches)-1); 
    end 
  | _ -> assert(false)

let rec subst (constr: constr) (boolvar: int) (substc: constr): constr =
  let subst_conn c =
    let subst_fun c = subst c boolvar substc in 
    let new_branches= List.map subst_fun c.branches in
    let new_wakeup_table = make_wakeup_table new_branches (Array.length c.wakeup_table) in 
    { c with branches= new_branches; wakeup_table= new_wakeup_table } in
  match constr with
  | BoolVar k -> 
      if k = boolvar then substc 
      else BoolVar k
  | Conjunction c -> Conjunction (subst_conn c)
  | Disjunction c -> Disjunction (subst_conn c)
  | UnivQuant q -> UnivQuant { q with subconstr= subst q.subconstr boolvar substc }
  | ExistQuant q -> ExistQuant { q with subconstr= subst q.subconstr boolvar substc }
  | c -> c

(*
let rec addvars (constr: constr) (free: int) (before: int) (after: int): constr= 
  let trans v= v+(if v>=free then before+after else before) in
  match constr with
  | BoolVar k -> BoolVar k 
  | BoolIntv b -> BoolIntv b
  | Unary c -> Unary { c with var1= trans c.var1 }
  | Binary c -> Binary { c with var21= trans c.var21; var22= trans c.var22 } 
  | Ternary c -> Ternary { c with var31= trans c.var31; var32= trans c.var32; var33= trans c.var33 }
  | Atomic a -> Atomic { a with aconstr= addvars a.aconstr free before after  }
  | Power c ->  Power { c with var01= trans c.var01;  var02= trans c.var02 }
  | SubSet c ->  SubSet { subset_var1= trans c.subset_var1; subset_var2= trans c.subset_var2 }
  | IsElement c -> IsElement { c with var= trans c.var }
  | PosSemDef p -> PosSemDef (Array.map (Array.map (function v -> trans v)) p)
  | Conjunction c -> 
      Conjunction { c with branches= List.map (function c -> addvars c free before after) c.branches; 
		           changed= c.changed; 
		           wakeup_table= Array.concat [Array.make (before+after) []; c.wakeup_table] }
  | Disjunction c -> 
      Disjunction { c with branches= List.map (function c -> addvars c free before after) c.branches; 
		           changed= c.changed;
                           wakeup_table= Array.concat [Array.make (before+after) []; c.wakeup_table] }
  | UnivQuant c -> 
      UnivQuant { c with subconstr= addvars c.subconstr free before after; 
		         outer_changed= List.map ((+) (before+after)) c.outer_changed;
		         inner_changed= List.map ((+) (before+after)) c.inner_changed }
  | ExistQuant c -> 
      UnivQuant { c with subconstr= addvars c.subconstr free before after; 
		         outer_changed= List.map ((+) (before+after)) c.outer_changed;
		         inner_changed= List.map ((+) (before+after)) c.inner_changed }
  | UnivQuantBound c -> 
      UnivQuantBound { 
        var_trans = Array.map trans c.var_trans;
        qbsubconstr= addvars c.qbsubconstr free before after }
  | Redundant c ->
      Redundant { original_constr = addvars c.original_constr free before after;
		  redundant_constr = addvars c.redundant_constr free before after }
*)
  
let switch_branch_off (constr: constr) (num: int): constr=
  (* note that this would be easier to implement if the branches of connectives would use an array instead of a list! *)
  let rec repl_branch (branches: constr list) (num: int) (b: bool): constr list =
    if num==0 then BoolIntv (if b then Bool_intv.true_intv else Bool_intv.false_intv)::(List.tl branches)
    else (List.hd branches)::repl_branch (List.tl branches) (num-1) b in
  assert(num>=0);
  let res= 
    match constr with 
      Conjunction c -> Conjunction { c with branches= repl_branch c.branches num true }
    | Disjunction c -> Disjunction { c with branches= repl_branch c.branches num false }
    |  _ -> assert(false) in 
  res

let build_connective (constr_list: constr list) (num_vars: int): connective =
  { branches= constr_list; 
    next_branch= 0; 
    bloated= true; 
    changed= Util.int_list 0 ((List.length constr_list)-1); 
    wakeup_table= make_wakeup_table constr_list num_vars;
    maxconj_pos= false;
    maxconj_neg= false
   }

let rec num_primitive (c: constr): int = 
  match c with 
    BoolVar _ | UserDefined _ | BoolIntv _ | Unary _ | Binary _ | Ternary _ | Power _ | SubSet _ | IsElement _ | PosSemDef _ -> 1
  | Atomic a -> num_primitive a.aconstr
  | Conjunction c -> List.fold_left (+) 0 (List.map num_primitive c.branches) 
  | Disjunction c -> List.fold_left (+) 0 (List.map num_primitive c.branches) 
  | UnivQuant c -> num_primitive c.subconstr 
  | ExistQuant c -> num_primitive c.subconstr 
  | Redundant r ->  (num_primitive r.original_constr)+(num_primitive r.redundant_constr)
  | UnivQuantBound c -> num_primitive c.qbsubconstr

let add_branch (conj: constr) (new_branch: constr): unit =
  match conj with
    Conjunction conn -> 
      let new_index= List.length conn.branches in
      conn.branches<- conn.branches @ [ new_branch ];
      conn.changed<- new_index::conn.changed;
      let add_branch_var v l =
	if contains_var new_branch v then 
	  new_index::l 
	else
	  l in
      conn.wakeup_table<- Array.mapi add_branch_var conn.wakeup_table 
  | _ -> assert(false)

let decorate (c: constr): unit= 
  let rec decorate_intern (under_atomic: bool) (c: constr): bool (* maxconj_pos *) * bool (* maxconj_neg *) =

    let rec unset_relax (pos: bool) (c: constr): unit = (* sets relax flags to false *)
      match c with 
	BoolVar _ | UserDefined _ | BoolIntv _ | Unary _ | Binary _ | Ternary _ | Power _ | SubSet _ | IsElement _ | PosSemDef _ -> ()
      | Atomic a -> unset_relax pos a.aconstr 
      | Conjunction c | Disjunction c -> 
	  if true then c.maxconj_pos<- false    
	  else c.maxconj_neg<- false;
      | UnivQuant c -> unset_relax pos c.subconstr 
      | ExistQuant c -> unset_relax pos c.subconstr 
      | Redundant r -> unset_relax pos r.original_constr (* ????? *)
      | UnivQuantBound c -> () (* ????? *) in
    (* end unset_relax *)

  (* begin decorate_intern *)
  match c with 
    BoolVar _ | UserDefined _ | BoolIntv _ | Unary _ | Binary _ | Ternary _ | Power _ | SubSet _ | IsElement _ | PosSemDef _ -> (true, true)
  | Atomic a -> decorate_intern true a.aconstr
  | Conjunction c -> 
      let rel= 
	if under_atomic then 
	  true 
	else
	  List.for_all (function (x,_) -> x) (List.map (decorate_intern under_atomic) c.branches) in
      if rel then List.iter (unset_relax true) c.branches;
      c.maxconj_pos<- rel;
      c.maxconj_neg<- false;
      (rel, false)
  | Disjunction c -> 
      let rel= 
	List.for_all (function (_,x) -> x) (List.map (decorate_intern under_atomic) c.branches) in
      if rel then List.iter (unset_relax false) c.branches;
      c.maxconj_pos<- false;
      c.maxconj_neg<- rel;      
      (false, rel)
  | UnivQuant c -> ignore(decorate_intern under_atomic c.subconstr); (false, false)
  | ExistQuant c -> ignore(decorate_intern under_atomic c.subconstr); (false, false)
  | Redundant r -> decorate_intern under_atomic r.original_constr (* ????? *)
  | UnivQuantBound c -> (true, true) (* ????? *) in
  (* end decorate_intern *)
  if !linear_relaxation then 
    ignore(decorate_intern false c)


