(* calculates the degree deg(f,B,0) where B is a box, f is an interval-computable function (box -> box) nonzero at \partial B
if 0\in f(\partial B), the degree is not defined and the algorithm would never end *)
open Box
open Interval
open Constr
open Parser_access;;



type orientation = int (* -1 or 1 *) 
type oriented_box = { box: box; orientation: orientation } (* structure *)

type sign = int (* -1/0/1 *)
type sign_vector = (sign array)

let header =
  print_endline "==================================================================";
  print_endline ("                         TopDeg 1.0");
  print_endline "                              by";
  print_endline "                        Peter Franek";
  print_endline "                   Czech Academy of Sciences";
  print_endline "with";
  print_endline "  Stefan Ratschan and Tomas Dzetkulic";
  print_endline "";
  print_endline "includes:";
  print_endline "  smathlib by Timothy Hickey and David Wittenberg";
  print_endline "==================================================================";
  print_newline() 

let rec ipow (a:int) (n:int):int =  (* integer power function *)
   assert(n>=0); (* defined only for nonnegative powers  *)
   if n=0 then 1 
   else 
     a * ipow a (n-1)

let print_sv (sv:sign_vector) = (* prints a sign-vector; for debugging *)  
  let l=Array.length sv in
  print_string "sign vector:";
  print_string " (";
  for i=0 to (l-1) do
    print_int(sv.(i));
    if i<l-1 then print_string ",";
  done;
  print_string ")\n"

let print_signed_box (sb:(oriented_box * sign_vector))=
   Box.print (fst sb).box; print_string " ";
   print_string " orientation="; print_int (fst sb).orientation; print_string " ";
   print_sv (snd sb);
   print_newline()

let rec print_bi(bi: (oriented_box * sign_vector) list) = 
	if bi!=[] then begin
           print_signed_box (List.hd bi);
	   print_bi (List.tl bi);
        end


let dimension (b: box):int = (* function for topological box dimension *)
   if Box.is_empty b then
      (-1) (* dimension of empty box *)
   else
      let sum=ref 0 in
      for i=0 to (Box.dim b)-1 do
	if (get_bound b i).lo<(get_bound b i).hi then 
		sum:=!sum+1;
      done;
      !sum

let box_to_sign_vector (b: box): sign_vector =
  assert(not (Box.is_empty b));
  let interval_to_sign (iv: interval): sign =
    if iv.lo>0.0 then 1
    else if iv.hi <0.0 then (-1) 
    else 0 in
  Array.init (Box.dim b) (function i -> interval_to_sign (get_bound b i)) (* probably the output of this function is Array of signs *)

let split (b: oriented_box): oriented_box * oriented_box =  (* splits an oriented box into two oriented boxes *)
  let var = widest_var b.box (Box.dim b.box) in
  let (b1, b2)= Box.bisect b.box var in
  ({ box= b1; orientation= b.orientation }, { box= b2; orientation= b.orientation })

let rec refineCov (f: box -> box) (l: oriented_box list): (oriented_box * sign_vector) list = (* refines a covering of \partial B by oriented boxes so that after for all
												box, there exists a function component with constant sign on that box *)
  match l with
    [] -> []
  | first::rest -> 
      let eval: sign_vector= box_to_sign_vector (f first.box) in
      let determined (sv: sign_vector): bool = (Util.array_exists (function s -> s<>0) eval)>=0 in
      if determined eval then begin
	(first, eval)::refineCov f rest
	end
      else begin
	let (b1, b2)= split first in
	refineCov f (b1::b2::rest)
      end

let project (sv: sign_vector) (i: int)= (* creates a sign vector one dimension smaller, where component i isignored *)
  assert(i>=0);
  let tilde_i (j:int):sign=
     if j<i then
	Array.get sv j
     else
	Array.get sv (j+1)
     in
  Array.init (Array.length sv-1) tilde_i

let calculate_oriented_boundary (b:oriented_box):oriented_box list= (* creates a list of oriented faces with the induced orientation *)
   let d=dimension (b.box) in
   let i_th_nonconstant_dim (i:int) (bx:box):int= (* returns the index of the i'th nonconstant dimension *)
      assert(i<dimension bx);
      let j= ref (-1) and ind=ref (-1) in
      while(!ind<i) do
	j:=!j+1;
	if (get_bound bx !j).lo<(get_bound bx !j).hi then ind:=!ind+1;
      done; 
      assert(!ind<d);
      !j in (* end of i_th_nonconstant_dim *)
   let upper_side (i:int) (b:oriented_box):oriented_box= (* finds the i'th nonconstant dimension *)
     let newbox=Box.init (Box.dim b.box) (function j-> (b.box).bounds.(j)) in
     let ind=i_th_nonconstant_dim i b.box in
     set_bound newbox ind (float_to_intv (get_bound b.box ind).hi); 
     {box=newbox; orientation=b.orientation*(ipow (-1) i)} in (* orientation of upper faces *)
   let lower_side (i:int) (b:oriented_box):oriented_box= (* finds the i'th nonconstant dimension *)
     let newbox=Box.init (Box.dim b.box) (function j-> (b.box).bounds.(j)) in
     let ind=i_th_nonconstant_dim i b.box in
     set_bound newbox ind (float_to_intv (get_bound b.box ind).lo); 
     {box=newbox; orientation=b.orientation*(ipow (-1) (i+1))} in (* orientation of the lower faces *)
   let newfaces_hi=Array.init d (function x-> upper_side x b)
   and newfaces_lo=Array.init d (function x-> lower_side x b) in
   List.concat [Array.to_list newfaces_lo; Array.to_list newfaces_hi] (* construction of oriented boundary *)

let split_along_hyperplane (ob:oriented_box) (ind:int) (r:float):(oriented_box*oriented_box)= (* splits an oriented box into two oriented subboxes, along the hyperplane x_ind = r *)
   assert(Interval.contains (get_bound (ob.box) ind) r);
   let b1=Box.copy ob.box in
   b1.bounds.(ind).hi <- r;
   ob.box.bounds.(ind).lo <- r; 
   (ob,{box=b1;orientation=ob.orientation})

let intersect (b1:box) (b2:box):(bool*int*float)= 
(* 	b1 is a box of dimension k-1, b2 a box of dimension k
	output is a triple (bool, int, float) 
	bool: true if they have a b1 dimensional intersection, 
	int: (-1 if b1\subset b2 or bool=false) or the index of the dimension in which they nontrivially intersect
	r: x_ind=r is the boundary of the intersection in the int'th dimension *)
   assert(dimension b2=dimension b1+1);
   let nontriv=ref true and i=ref 0 and n=Box.dim b1 and ind= ref (-1) and r=ref 0. in
   while ((!i<n) && (!nontriv)) do
      if b1.bounds.(!i).lo > b2.bounds.(!i).hi || b1.bounds.(!i).hi < b2.bounds.(!i).lo || (b1.bounds.(!i).lo < b1.bounds.(!i).hi && (b1.bounds.(!i).lo=b2.bounds.(!i).hi || b1.bounds.(!i).hi=b2.bounds.(!i).lo || b2.bounds.(!i).lo=b2.bounds.(!i).hi)) then
	nontriv:=false (* no b1-dimensional intersection *)
      else if b1.bounds.(!i).lo < b1.bounds.(!i).hi then (* nontrivial dimension *)
	 if b2.bounds.(!i).lo > b1.bounds.(!i).lo  then begin (* b2 splits b1 *)
	     ind:=!i;
	     r:=b2.bounds.(!i).lo
	 end
	 else if b2.bounds.(!i).hi < b1.bounds.(!i).hi then begin
	     ind:=!i;
	     r:=b2.bounds.(!i).hi;
	 end;
      i:=!i+1;
   done;
   (!nontriv, !ind, !r) (* returns the triple *)

let rec sign_box (ob:oriented_box) (other_boxes:(oriented_box*sign_vector) list) (l:int):((oriented_box*sign_vector) list)=
	(* returns the oriented_box*sign_vector list of ob boundary, inherited from other_boxes *)
   if other_boxes=[] then []
      else begin
      let (non_triv, ind, r)=intersect (ob.box) ((fst (List.hd other_boxes)).box) in 
      (* non_triv=codimension 0 intersection exists, ind,r=  one has to split the box in x_ind=r  *)
      if (not non_triv) then 
	 sign_box ob (List.tl other_boxes) l (* if intersection not found, continue with the rest of other_boxes *)
      else if ind=(-1) (* symbol ob \subset List.hd other_boxes *) then
	 [(ob,project (snd (List.hd other_boxes)) l)]
	   else begin
	      let (ob1,ob2)=split_along_hyperplane ob ind r in (* splits ob in dimension ind along x_ind=r *)
	      List.append (sign_box ob1 other_boxes l) (sign_box ob2 other_boxes l)
	   end
      end

let rec sign_headboundary (headboundary:oriented_box list) (other_boxes: (oriented_box * sign_vector) list) (l:int):(oriented_box*sign_vector) list=
   if headboundary=[] then
     []
   else begin
     let sb = sign_box (List.hd headboundary) other_boxes l in (* returns an empty box or a list of signed oriented boxes, splitting of (List.hd headboundary) *)
     List.append sb (sign_headboundary (List.tl headboundary) other_boxes l) (* recursively creates the signed oriented boundary of the headboundary box *)
   end 

let strategy_choice (boundary_info: (oriented_box * sign_vector) list) (dim:int):(int * int)= (* calculates the position and sign sich that the set of elements in boundary_info with sign-vector
												having the sign on the particular component has minimal cardinality  *)
      let deb=false in
      let is_rel ((ob, sv): (oriented_box * sign_vector)) (pos: int) (sign:int)= (sv.(pos)= sign) in 
      let count_relevant_plus (bound: (oriented_box * sign_vector) list) (pos:int) : int = 
        let (sel_plus, other_plus) = List.partition (function x -> is_rel x pos 1) boundary_info in
        List.length sel_plus in

      let count_relevant_minus (bound: (oriented_box * sign_vector) list) (pos:int) : int= 
        let (sel_minus, other_minus) = List.partition (function x -> is_rel x pos (-1)) boundary_info in
        List.length sel_minus in
     
      let count_boxes_plus = Array.init (dim+1) (function x-> count_relevant_plus boundary_info x)
      and count_boxes_minus = Array.init (dim+1) (function x-> count_relevant_minus boundary_info x) in
      let m=ref count_boxes_plus.(0) in
      let l=ref 0 and s= ref 1 in
      for i=0 to (dim) do 
        let current_length_plus=count_boxes_plus.(i) in
        let current_length_minus=count_boxes_minus.(i) in
        if deb then begin print_string "current_length_minus="; print_int current_length_minus; print_newline();  end;
        if current_length_minus < !m then begin
           m:=current_length_minus;
           l:=i;
           s:= (-1);
	end;
      done;
      (!l, !s)

let rec deg (boundary_info: (oriented_box * sign_vector) list): int = (* for technical reasons, it returns twice the degree *)
  let strat=true in
  if boundary_info=[] then
     0
  else begin
    let k = dimension (fst (List.hd boundary_info)).box in (* dimension we are working in *)
    assert(List.for_all (function x -> ((dimension (fst x).box)=k) && (Array.length (snd x)=k+1)) boundary_info); (* check the correctness of input *)
    assert (k>=0);
    if k=0 then begin 
      if List.length(boundary_info)=1 then 
        let p1 = List.hd boundary_info in
        assert(not ((snd p1).(0) = 0)); (* the sign-vectors are determined *)  
        ((fst p1).orientation)*((snd p1).(0))
      else  
	begin
        let p1 = List.hd boundary_info in
        assert(not ((snd p1).(0) = 0)); (* the sign-vectors are determined *)  
        ((fst p1).orientation)*((snd p1).(0))+deg (List.tl boundary_info)
      end  
    end else begin (* k is not 0 *)
      let (l,s)= if strat then strategy_choice boundary_info k else (0,-1) in 
      let is_relevant ((ob, sv): (oriented_box * sign_vector))= (sv.(l)= s) in 
      let (selected_boxes, other_boxes) = List.partition is_relevant boundary_info in (* selected boxes are those with the particular sign-vector *)
      let rec faces (selected_boxes:(oriented_box * sign_vector) list) (other_boxes: (oriented_box * sign_vector) list): (oriented_box*sign_vector) list=  (* returns the oriented signed boundary of the selected boxes *)
	  if (List.length selected_boxes)=0 then
		[]
	  else begin
	    let headboundary=calculate_oriented_boundary (fst (List.hd selected_boxes)) in (* calculates oriented boundary of the head *)
            let headfaces=sign_headboundary headboundary other_boxes l in (* possibly splits the boundary boxes having k-1 dimensional intersection with some selected boxes and calculates its sign vectors *)
            List.append headfaces (faces (List.tl selected_boxes) other_boxes) (* concatenates the faces of the head and of the tail *)
	    end in
	(ipow (-1) l)*s*deg (faces selected_boxes other_boxes) (* calles the degree function with smaller dimensional faces *)
    end
  end

(*************************** MAIN: ********************************)
let main =
header;
print_string "Variables: "; flush_all ();
let var = parse_vars () in 
let reader (a:int)=
(parse_term var) in
print_string "Term: "; flush_all ();
let terms = Array.init (List.length var) reader in
(* print_constr c; *)
print_string "Box: "; flush_all ();
let b = parse_box () in 
if (not ((dim b)=(List.length var))) then begin
  print_endline "Number of variables different from bounding box dimension!"; exit(1)
end;
(* print_endline "Start "; *)
let ob=calculate_oriented_boundary {box=b;orientation=1} in (* list of oriented boundary faces *)
(* print_endline "Boundary computed "; *)
let evaluate (bp:box):box=
let evaluate_single (idx:int):interval=
let (v,t)=terms.(idx) in
(Ieval.ieval_term t bp v) in
(Box.init (List.length var) evaluate_single) in
let boundary_info=refineCov evaluate ob in (* possible refinement of the boundary boxes *)
(* print_endline "Boundary info computed"; *)
print_string "degree =  ";
print_int (deg boundary_info/2); (* printing the degree *) 
print_newline();

