open Interval
open Constr
open Box
open Parser
open Lexer 

type witness = (float array * bool (* verified? *) ) option 

let trace= false
let trace_atomic = false

let flag_remove = ref true   
let flag_shortcut = ref false
let flag_sample = ref true
let flag_sample_approximate = ref false
let flag_marks = ref true

let deduce_atomic_ctr = ref 0
let improve_atomic_ctr = ref 0
let deduce_primitive_ctr= ref 0
let mark_cons_ctr= ref 0

let prop_threshold = ref 0.01 (* used for finite intervals 
For infinite intervals we have the following situations:
   - change from two-sided infinite to one-sided 
   - change from infinite to finite 
   - change of finite bound of infinite interval

Currently, the first two result in wakeups, the last one not
*)

let change_below_threshold= true

(* definitions for fixpoint computation *)
type cons_constr={mutable cstr: constr ; mutable flag: bool}

(* not in constr.ml, because there variables are hidden *)
let constr_varlist (constr : constr) (bound: int): variable list=
  (* all variables, that are smaller than bound; this bound has been introduced to avoid calls to Util.union with long lists *)
  let rec varlist (constr: constr) =
  match constr with
    BoolVar _   -> [] |
    BoolIntv _   -> [] |
    Ternary p1	 -> List.filter (function v -> v<bound) [p1.var31;p1.var32;p1.var33]|
    Binary p2 	 -> List.filter (function v -> v<bound) [p2.var21;p2.var22]|
    IsElement p3 -> if p3.var<bound then [p3.var] else [] |
    Power p4  	 -> List.filter (function v -> v<bound) [p4.var01;p4.var02] |
    Unary p5 	 -> if p5.var1<bound then [p5.var1] else [] |
    SubSet c 	 -> List.filter (function v -> v<bound) [c.subset_var1; c.subset_var2]|
    PosSemDef p -> let pack_list v = [ v ] in
                   let varlist a = Array.fold_left (@) [] (Array.map pack_list a) in
                   Array.fold_left (@) [] (Array.map varlist p) |
    Conjunction p6 -> List.fold_left Util.union [] (List.map varlist p6.branches)  |
    Disjunction p7 -> List.fold_left Util.union [] (List.map varlist p7.branches) |
    Atomic q -> varlist q.aconstr| 
    ExistQuant q -> varlist q.subconstr |
    UnivQuant q -> varlist q.subconstr
  | Redundant c -> Util.union (varlist c.original_constr) (varlist c.redundant_constr)
  | UnivQuantBound c -> varlist c.qbsubconstr 
  | UserDefined (_, contains_var) -> 
      let rec varlist_start i =
	if i>=bound then []
	else if contains_var i then i::(varlist_start (i+1))
	else (varlist_start (i+1)) in
      varlist_start 0 in
  varlist constr 

let rec sample_able (constr: constr) (neg: bool): bool =
  (* can we compute a satisfying (non if 'neg') sample of 'constr' using local search? *)
  (*
  match constr with
    Atomic _ -> true
  | Conjunction l -> not neg
  | Disjunction l -> neg
  | _ -> false
  I do not remember the exact motivation for this. Try to sample always*) 
  true


(**************************************************************************************************************
	                 pruning atomic constraints
 **************************************************************************************************************)
let rec pruneBoolIntv ({ Bool_intv.lo= lo; Bool_intv.hi= hi }: Bool_intv.bintv) (box: box) (outer: bool): witness = 
  if (outer && (not hi)) || ((not outer) && lo) then
    set_empty box;
  None 
and pruneTernary (t:primTernary) (box: box): witness =
  deduce_primitive_ctr:= !deduce_primitive_ctr + 1;
  begin
    match t.predicate3 with
      Addition -> 
	if not (narrow_add (get_bound box t.var31 )  (get_bound box t.var32 ) (get_bound box t.var33 )) then
          set_empty box 
    | Subtraction -> 
	if not (narrow_sub (get_bound box t.var31 )  (get_bound box t.var32 ) (get_bound box t.var33)) then
          set_empty box  
    | Multiplication -> 
	if not (narrow_mul (get_bound box t.var31 )  (get_bound box t.var32 ) (get_bound box t.var33)) then
	  set_empty box;
  end;
  None

 and prunePower (t:primPower) (box:box): witness = 
  if (t.n mod 2)=0 then begin
    if not (narrow_pow_even (get_bound box t.var01 ) (float_to_intv (float_of_int t.n)) (get_bound box t.var02 )) then
      set_empty box
  end else begin
    if not (narrow_pow_odd (get_bound box t.var01 ) (float_to_intv (float_of_int t.n)) (get_bound box t.var02 )) then
      set_empty box;
  end;
  None

 and pruneUnary (t:primUnary) (box: box) (outer: bool) =
  assert(t.var1<(dim box));
  deduce_primitive_ctr:= !deduce_primitive_ctr + 1;
  let p= if outer then t.predicate1 else neg_unary t.predicate1 in 
  begin
    match p with
      Equal -> if not (narrow_eq (get_bound box t.var1)) then set_empty box; |
      LessEqual ->if not (narrow_le (get_bound box t.var1)) then set_empty box; | 
      Less -> if not (narrow_lt (get_bound box t.var1)) then set_empty box; |
      NotEqual -> if not (narrow_ne (get_bound box t.var1)) then set_empty box; |
      Greater -> if not ( narrow_gt (get_bound box t.var1)) then set_empty box; |
      GreaterEqual -> if not (narrow_ge (get_bound box t.var1)) then set_empty box; 
  end;
  None

and pruneBinary (t:primBinary) (box: box): witness =
  deduce_primitive_ctr:= !deduce_primitive_ctr + 1;
  begin
    match t.predicate2 with
      Square -> if not (narrow_square (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | Sin -> if not (narrow_sin  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | Cos -> if not (narrow_cos  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | Tan -> if not (narrow_tan  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | ArcSin -> if not (narrow_asin  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | ArcCos -> if not (narrow_acos  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | ArcTan -> if not (narrow_atan  (get_bound box t.var21) (get_bound box t.var22)) then set_empty box; 
    | Exp ->if not (narrow_exp (get_bound box t.var21 ) (get_bound box t.var22)) then set_empty box; 
    | Log ->if not (narrow_log (get_bound box t.var21 ) (get_bound box t.var22)) then set_empty box;
    | UMinus->if not (narrow_u_minus (get_bound box t.var21 ) (get_bound box t.var22)) then set_empty box;
  end;
  None
and pruneIsElem (t:isElement) (box: box): witness = 
  deduce_primitive_ctr:= !deduce_primitive_ctr + 1;
  let iv= (Interval.intersect t.interval (get_bound box t.var)) in 
    if iv.Interval.lo<=iv.Interval.hi then set_bound box t.var iv
    else set_empty box;
  None
and pruneSubSet (t:subset) (box: box): witness =
  deduce_primitive_ctr:= !deduce_primitive_ctr + 1;
  let iv= Interval.intersect (get_bound box t.subset_var1) (get_bound box t.subset_var2) in 
  if Interval.is_empty iv then set_empty box
  else set_bound box t.subset_var1 iv;
  None
and prunePosSemDef (p: posSemDef) (box: box) (outer: bool) (sample: bool): witness=
  (* sampling currently only implemented for inner pruning! *)
  if (classify_float (width box))=FP_normal then  (* SDP solver cannot cope with infinite bounds *) 
    let interval_array a = Array.map (function v -> get_bound box v) a in 
    let interval_matrix = Array.map interval_array p in begin
(*      let print_matrix m = 
	for i = 0 to (Array.length m)-1 do
	  for j= 0 to (Array.length m.(i))-1 do
	    Interval.print m.(i).(j)
	  done;
	  print_newline ()
	done in *)
(*      print_matrix interval_matrix; print_endline (if outer then "outer" else "inner"); flush stdout;   *)
      if outer then (begin
(*	print_matrix interval_matrix; print_endline (if outer then "outer" else "inner"); flush stdout;     *)
	let record_factor = 0.1 in (* only record changes greater than this fraction *)
	for i= 0 to (Array.length p)-1 do
	  for j= i to (Array.length p)-1 do 
	    if not ((is_empty box) || is_thin (get_bound box p.(i).(j))) then begin
(* 	      print_int i; print_int j; print_endline "false";  *)
	      match Sdp.sdp interval_matrix (i,j) false with
		(Sdp.SDPSuccess, lower) -> 
		  if lower>box.bounds.(p.(i).(j)).Interval.lo +. record_factor *. (Interval.width (get_bound box p.(i).(j))) then begin
(* 		    print_string "l: "; print_int p.(i).(j); print_string " "; print_float lower; print_newline ();*)
		    box.bounds.(p.(i).(j)).Interval.lo<- min lower (get_bound box p.(i).(j)).Interval.hi;
		    box.bounds.(p.(j).(i)).Interval.lo<- min lower (get_bound box p.(j).(i)).Interval.hi;
		  end |
		  (Sdp.SDPNoSuccess, _) -> () |
		  (Sdp.SDPUnbounded, _) -> () (* assert(false) *) |
		  (Sdp.SDPNotFeasible, _) -> set_empty box
	    end;
	    if not ((is_empty box) || is_thin (get_bound box p.(i).(j))) then begin
(*	      print_int i; print_int j; print_endline "";  *)
	      match Sdp.sdp interval_matrix (i,j) true with
		(Sdp.SDPSuccess, upper) -> 
		  if upper<box.bounds.(p.(i).(j)).Interval.hi  -. record_factor *. (Interval.width (get_bound box p.(i).(j))) then begin
(*		    print_string "u: "; print_int p.(i).(j); print_string " "; print_float upper; print_newline ();  *)
		    box.bounds.(p.(i).(j)).Interval.hi<- max upper (get_bound box p.(i).(j)).Interval.lo;
		    box.bounds.(p.(j).(i)).Interval.hi<- max upper (get_bound box p.(j).(i)).Interval.lo;
		  end |
		  (Sdp.SDPNoSuccess, _) -> () |
		  (Sdp.SDPUnbounded, _) -> () (* assert(false) *) |
		  (Sdp.SDPNotFeasible, _) -> set_empty box
	    end;
	    assert((is_empty box) || box.bounds.(p.(i).(j)).Interval.lo<=box.bounds.(p.(i).(j)).Interval.hi);
	  done
	done;
(* 	print_string "to: "; print box; print_endline ""; flush stdout; *)
	None
      end: witness) else (begin (* inner pruning: go through all corners and show whether they are feasible, if yes, by convexity, box is feasible *)
	if !flag_sample && sample then begin
	  let array_make i = Array.make (Array.length p) 0.0 in
	  let f= Array.init (Array.length p) array_make in
	  let witness_len = dim box in
	  if Sdp.sdpx interval_matrix f = Sdp.SDPSuccess then set_empty box;
	  if is_empty box then 
	    let witness= Array.make witness_len 0.0 in begin
	      for i= 0 to (Array.length p)-1 do
		for j= 0 to (Array.length p)-1 do
		  witness.(p.(i).(j))<- f.(i).(j)
		done
	      done; 
	      Some (witness, false)
	    end 
	  else None
	end else begin
	  let bloat_factor = 1.1
	  and (corner: int array array) = Array.make_matrix (Array.length p) (Array.length p) (-1)
	  and next (corner: int array array) = 
	    let i= ref 0 in
	    while !i<(Array.length corner)*(Array.length corner) && corner.(!i/(Array.length corner)).(!i mod (Array.length corner))>0 do
	      corner.(!i/(Array.length corner)).(!i mod (Array.length corner))<- -1;
	      i:= !i+1
	    done;
	    if !i<(Array.length corner)*(Array.length corner) then begin
	      corner.(!i/(Array.length corner)).(!i mod (Array.length corner))<- 1;
	      false
	    end else begin
	      true
	    end 
	  and is_symmetric (corner: int array array): bool =
	    let symm= ref true 
	    and i= ref 0 
	    and j= ref 0 in begin
	      while !i<Array.length p && !symm do
		j:= 0;
		while !j < !i && !symm do
		  if not (corner.(!i).(!j)=corner.(!j).(!i)) then
		    symm:= false;
		  j:= !j+1
		done;
		i:= !i+1
	      done;
	      !symm
	    end in
	  let corner_reached result (corner: int array array) = 
	    let reached= ref true 
	    and i= ref 0 
	    and j= ref 0 in begin
	      while !i<Array.length p && !reached do
	      j:= 0;
		while !j<Array.length p && !reached do
(*		  print_int !i; print_int !j; print_string " "; print_float result.(!i).(!j); print_string " "; 
                  if corner.(p.(!i).(!j))>0 then print_float (get_bound box p.(!i).(!j)).hi
                  else print_float (get_bound box p.(!i).(!j)).lo;
                  print_endline ""; *)
		  if (not (is_thin (get_bound box p.(!i).(!j)))) && 
		    ((corner.(!i).(!j)>0 && result.(!i).(!j)<(get_bound box p.(!i).(!j)).Interval.hi) ||
	            (corner.(!i).(!j)<0 && result.(!i).(!j)>(get_bound box p.(!i).(!j)).Interval.lo)) then
		    reached:= false;
		  j:= !j+1
		done;
		i:= !i+1
	      done;
(*	    flush stdout; *)
	      !reached
	    end
	  and bloat iv = begin
	    { Interval.lo= iv.Interval.hi -. (iv.Interval.hi -. iv.Interval.lo) *. bloat_factor;  
	      Interval.hi= iv.Interval.lo +. (iv.Interval.hi -. iv.Interval.lo)*. bloat_factor }
	  end
	  and finished = ref false 
	  and full_prune = ref true in begin
	    for i= 0 to (Array.length p)-1 do
	      for j= 0 to (Array.length p)-1 do
		interval_matrix.(i).(j)<- bloat interval_matrix.(i).(j) (* we optimize in a slightly larger matrix, so that we can return a result even if SDP solver returns sub-optimum *)
	      done
	    done;
	    while ((not !finished) && !full_prune) do 
(* 	      Array.iter (Array.iter print_int) corner; print_newline (); flush stdout;  *)
	      if is_symmetric corner then begin (* has to optimize corners that correspond to symmetric elements in the same direction *)
		let objective = Array.map (Array.map float_of_int) corner in 
		match Sdp.sdpx interval_matrix objective with
		  Sdp.SDPSuccess -> if not (corner_reached objective corner) then full_prune:= false |
		  Sdp.SDPNoSuccess -> full_prune:= false |
		  Sdp.SDPUnbounded -> full_prune:= false; (* assert(false) *) |
		  Sdp.SDPNotFeasible -> full_prune:= false 
	      end;
	      finished:= next corner;
	    done;
	    if !full_prune then begin
(*	      print_endline "Inner success"; flush stdout; *)
	      set_empty box;
	    end;
	  end;
	  None
	end
      end: witness);
    end else
    None


      
  

(**************************************************************************************************************
	                 Pruning atomic constraints
 **************************************************************************************************************)
and pruneAtomic (atomic: atomic) (box: box) (outer: bool) (changed: Intset.t) (sample: bool): witness =
(*  if outer then begin
    print_string "outer-in: "; 
    print box;  
  end;*)
  let improved= ref false 
  and witness= ref None in
  begin
    let aconstr = 
      begin
	match atomic.aconstr with 
	  Conjunction c -> c
	| _ -> assert(false)
      end in 
    deduce_atomic_ctr:= !deduce_atomic_ctr+1;
    if outer then begin
      if not ((dim atomic.obound)=0 || finite_width (get_bound atomic.obound 0)) then  (* assumption: narrowing operator result in finite bounds as soon as applied *)
	Intset.add changed  (Util.int_list (dim box) ((dim box)+(dim atomic.obound)-1));
    end else begin
      if not ((dim atomic.obound)=0 || finite_width (get_bound atomic.ibound 0)) then 
	Intset.add changed  (Util.int_list (dim box) ((dim box)+(dim atomic.ibound)-1));
    end;
    let p= cart_prod2 box (if outer then atomic.obound else atomic.ibound) in  
    ignore(pruneConjunction aconstr p outer changed sample);
    if is_empty p then begin 
      set_empty box;
      improved:= true;
    end else begin
      for i= 0 to (dim p)-1 do
	if i<(dim box) then begin
	    if not ((get_bound box i)=(get_bound p i)) then begin
	      assert(Interval.subset (get_bound p i) (get_bound box i));
	      improved:= true;
	      set_bound box i (get_bound p i)
	    end 
	end else begin
	  if outer then begin
	    if not ((get_bound atomic.obound (i-(dim box)))=(get_bound p i)) then 
	      set_bound atomic.obound (i-(dim box)) (get_bound p i)
	  end else begin
	    if not ((get_bound atomic.ibound (i-(dim box)))=(get_bound p i)) then 
	      set_bound atomic.ibound (i-(dim box)) (get_bound p i)
	  end
	end;
      done
    end;
  end;
  if !improved && trace_atomic then begin 
(*      Box.print box; *)
(*      Box.print p; *)
    print_endline "--------------------------------------------------------------"; flush_all ();
(*      print box;  *)
    print_bconstr (Atomic atomic, box); 
    if outer then print_endline "outer" else print_endline "inner";
    print_endline "--------------------------------------------------------------"; flush_all ()
  end;
  if !improved then improve_atomic_ctr:= !improve_atomic_ctr+1;
  !witness 

    
(**************************************************************************************************************
	                 pruning conjunctions
 **************************************************************************************************************)
 
and pruneConjunction ({ branches= t; next_branch= _; changed= changed_branches; wakeup_table= wakeup_table; 
                        maxconj_pos= rel_pos; maxconj_neg= rel_neg }: connective) (box: box) (outer: bool) (changed: Intset.t) (sample: bool): witness=  (* algorithm from new paper, without changed variable table *)
  assert(List.for_all (function c -> Constr.is_consistent c box) t);
  let new_changed = Intset.copy changed in
  let contr_table = 
    let prune_branch c b=
      prune_rec c b outer new_changed sample in
    let contr_list= List.map prune_branch t in
    Array.of_list contr_list 
  and contr_vars = 
    let l= (List.map (function c -> (constr_varlist c (dim box))) t) in
    (Array.of_list l: int list array) in
  let wakeup_table= Array.init (Array.length wakeup_table) (function i -> wakeup_table.(i)) in
  let is_branch i = i<List.length t in
  let contains_var i v = 
    (not (is_branch i)) || (Constr.contains_var (List.nth t i) v) in
  let queue = Array.make (Array.length contr_table) (-1) (* empty, -1=not enqueued, length=last queue element, otherwise next queue element *)
  and queue_start= ref ((Array.length contr_table))         (* (length queue) if queue is empty *)
  and queue_end= ref ((Array.length contr_table)) in
(*  let print_queue () =
    let p= ref !queue_start in
    print_string "Queue: ";
    while !p<(Array.length queue) do
      print_int !p; print_char ' ';
      p:= queue.(!p);
    done;
    print_newline () in *)
  let assert_consistent () = begin
    assert(0<=(!queue_start) && !queue_start<=(Array.length queue));
    assert(0<=(!queue_end) && (!queue_end<(Array.length queue) || !queue_end = !queue_start));
    for i= 0 to (Array.length queue)-1 do 
      if i=(!queue_end) && !queue_start<(Array.length queue) then
	assert(queue.(i)=(Array.length queue))
      else 
	assert(-1<=queue.(i) && queue.(i)<(Array.length queue))
    done
  end in 
  let queue_full () = begin
    for i= 0 to ((Array.length queue)-1) do queue.(i)<- i+1 done;
    queue_start:= 0;
    queue_end:= (Array.length queue)-1
  end
  and queue_empty () = !queue_start=(Array.length queue) in  
  let enqueue i = begin (* at the end *)
    assert(i<Array.length queue);
    if queue.(i)<0 then begin (* not already enqueued *)
(*      print_string "+"; print_int i; print_newline ();*)
      if queue_empty () then begin
	queue_start:= i;
	queue_end:= i;
	queue.(i)<- Array.length queue
      end else begin
	queue.(i)<- Array.length queue;
	queue.(!queue_end)<-i;
        queue_end:= i
      end
    end;
    assert( assert_consistent (); true)
  end
  and dequeue () = begin (* from the start *)
    assert(not (queue_empty ()));
    let next= queue.(!queue_start) in 
    queue.(!queue_start)<- -1; 
    queue_start:= next;
    assert( assert_consistent (); true)
  end in
  let fixpoint_loop () = (* separate function for profiling *) begin
    assert(Array.length(wakeup_table)<=(dim box));
    assert(Array.length(contr_table)=Array.length(queue));
    if !flag_marks then begin (* initialize queue *)
      List.iter enqueue changed_branches; (* all changed branches *)
      for v= 0 to (dim box)-1 do
	if Intset.is_element new_changed v then 
	  if v<Array.length wakeup_table then begin
	    List.iter enqueue wakeup_table.(v); (* all branches  that contain a changed variable *)
	  end;
      done;
    end else 
      queue_full (); (* initialize queue with all elements *)
    let witness= ref None
    and oldbox= copy box in 
    let wakeup v = 
      let old_iv= get_bound oldbox v 
      and new_iv= get_bound box v in
      if new_iv<>old_iv then begin
	if (!prop_threshold=0.) || 
	   (old_iv.lo=neg_infinity && old_iv.hi=infinity && (new_iv.lo<>neg_infinity || new_iv.hi<>infinity)) ||
           (Interval.diff_width_rel old_iv new_iv !prop_threshold) then begin  
	  List.iter enqueue wakeup_table.(v);  (* this does not re-enqueue constraint itself, since we dequeue it later *)
	  set_bound oldbox v (Interval.copy new_iv);
	  Intset.add_el changed v;
	end else begin
	  if (not change_below_threshold) then set_bound box v (Interval.copy (get_bound oldbox v))
	end
      end (* end wakeup *) in
    (* begin fixpoint_loop *)
    while not ((is_empty box) || queue_empty ()) do 
      witness:= contr_table.(!queue_start) box;
      if trace then begin 
	for i= 0 to (Box.dim box)-1 do
	  if contains_var !queue_start i then begin
	    print_char 'v'; print_int i; 
	    Interval.print (get_bound box i); 
	    if (get_bound box i)<>(get_bound oldbox i) then begin
	      print_char '*';
              if (get_bound box i).lo>(get_bound oldbox i).lo then print_char 'l';
   	      if (get_bound box i).hi<(get_bound oldbox i).hi then print_char 'u'; 
	    end;
	    print_char ' ';
	  end
	done;
(*	Box.print *)
	if is_branch !queue_start then 
	  print_constr (List.nth t !queue_start)
	else
	  print_endline "Relaxation Contractor";
      end ;
      if not (is_empty box) then begin
	List.iter wakeup contr_vars.(!queue_start);
	dequeue ()
      end
    done;
(*     if trace then begin Box.print box ; print_endline "c"; flush_all () end; *)
    !witness
  end in
  let witness= fixpoint_loop () in
  witness


(**************************************************************************************************************
	                 pruning non-conjunctions
 **************************************************************************************************************)

and pruneDisjunction (l: connective) (b: box) (outer: bool) (changed: Intset.t) (sample: bool): witness = 
(* if not outer then prune wrt. the negation of the sub-constraints; return an new equivalent list of constraints *)
(*  let is_atomic c = match c with Atomic _ -> true | _ -> false in *)
(*  print_newline "prune_disj"; print_box b; print_newline (); *)
  let i = ref 0 
  and new_bloated = ref false
  and b_old= copy b in 
  set_empty b;
  while !i<List.length l.branches do (* !flag_shortcut???*)
    let c= List.nth l.branches !i
    and b1= copy b_old 
    and new_changed= Intset.copy changed in 
    if l.bloated then Intset.add new_changed (Util.int_list 0 ((dim b_old)-1));
    ignore (prune_rec c b1 outer new_changed false);
    if (is_empty b1 && !flag_remove) then 
      connective_remove l !i 
    else begin
      if !i>0 && not (b1=b_old) then new_bloated:= true;
	union_d b b1;
      i:= !i+1
    end
  done;
  l.bloated<- !new_bloated;
  None
   
and pruneUniv (q: quantification) (b: box) (outer: bool) (changed: Intset.t) (sample: bool): witness =
(* if not outer then prune wrt. the negation of the sub-constraints *)
  let bq= cart_prod2 b q.qbound in begin
    let new_changed = Intset.copy changed in
    Intset.add new_changed (if outer then q.outer_changed else q.inner_changed);
    let witness= prune_rec q.subconstr bq outer new_changed (sample || (dim b)=0) in
    if is_empty bq then set_empty b 
    else begin
      let i= ref 0 in begin
	i:= (dim b);
	while !i<(dim bq) && not (is_empty b) do 
	  if not ((get_bound q.qbound (!i-(dim b)))=(get_bound bq !i)) then begin
	    set_empty b;
	  end;
	  i:= !i+1;
	done;
      end;
      if not (is_empty b) then b.bounds<-(proj bq 0 (dim b)).bounds
    end;
    witness
  end;
 (*
  if !flag_marks then 
    if outer then q.outer_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1)
    else q.inner_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1);
*)

and pruneUnivQuantBound (q: quantbound) (b: box) (outer: bool) (changed: Intset.t) (sample: bool): witness =
  let bq= 
    let newbounds= 
      let initfun i = Interval.float_to_intv (Interval.midpoint (get_bound b q.var_trans.(i))) in
      Box.init (Array.length q.var_trans) initfun in
    cart_prod2 b newbounds in 
  let new_changed = Intset.copy changed in
  for i= (dim b) to (dim b)+(Array.length q.var_trans)-1 do
    if Intset.is_element changed q.var_trans.(i-(dim b)) then
      Intset.add new_changed [i]
  done;
  let witness= prune_rec q.qbsubconstr bq outer new_changed (sample || (dim b)=0) in
  if is_empty bq then set_empty b
  else b.bounds<-(proj bq 0 (dim b)).bounds;
  clear q.qbsubconstr; (* since we have taken a sample, we cannot use the new bounds further *)
  witness
 (*
  if !flag_marks then 
    if outer then q.outer_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1)
    else q.inner_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1);
*)

and pruneExist (q: quantification) (b: box) (outer: bool) (changed: Intset.t) (sample: bool): witness =
(*  print b;
  print_endline (if outer then "pruneExist outer" else "pruneExist inner"); *)
(* if not outer then prune wrt. the negation of the sub-constraints *)
  let p= cart_prod2 b q.qbound in
(*    if (not outer) then begin
   assert(Constr.is_consistent q.subconstr p);
   match q.subconstr with
   Conjunction c -> pruneConjunction c p outer (changed @ q.inner_changed) |
   _ -> assert(false) end
   else  *)
  let new_changed = Intset.copy changed in
  Intset.add new_changed (if outer then q.outer_changed else q.inner_changed);
  let witness= prune_rec q.subconstr p outer new_changed false in
  if outer then
    q.outer_changed<- []
  else 
    q.inner_changed<- [];
  if is_empty p then set_empty b
  else begin
    for i= 0 to (dim p)-1 do
      if i<(dim b) then begin
	if not ((get_bound b i)=(get_bound p i)) then 
	  set_bound b i (get_bound p i)
      end else begin
	if not ((get_bound q.qbound (i-(dim b)))=(get_bound p i)) then begin
	    (* if (not prim) then *) set_bound q.qbound (i-(dim b)) (get_bound p i);
	  if !flag_marks then 
	    if outer then q.inner_changed<- Util.union [ i ] q.inner_changed 
	    else q.outer_changed<- Util.union [ i ] q.outer_changed
	end
      end
    done
  end;
  witness
(*
  if !flag_marks then 
    if outer then 
      q.outer_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1)
    else 
      q.inner_changed<- Util.int_list (dim b) ((dim b)+(dim q.qbound)-1);
*)


(**************************************************************************************************************
	                 overall pruning
 **************************************************************************************************************)
and prune_rec (constr: constr) (box: box) (outer: bool) (changed: Intset.t) (sample: bool): witness =
  assert(Constr.is_consistent constr box);
(*  assert(Util.unique changed);*)
(*  assert(List.for_all (function v -> v>=0) changed); *)
  if (is_empty box) (* || (!flag_marks && (consistent_marks constr (dim box) outer cons)) *) then begin
(*    if not (is_empty box) then mark_cons_ctr:= !mark_cons_ctr+1 *)
    None
  end else if sample && !flag_sample && (sample_able constr outer) then begin
    let witness= compute_sample constr box outer in
    if Util.is_something witness then begin
      if trace || trace_atomic then print_endline "sample found";
      set_empty box
    end;
    witness
  end else
    match constr with
      Ternary p1 -> pruneTernary p1 box 
    | Binary p2 -> pruneBinary p2 box 
    | IsElement p3 -> pruneIsElem p3 box 
    | Power p4 -> prunePower p4 box 
    | Unary p5 -> pruneUnary p5 box outer (* only unary constraints are negated !!!!!!!!!! *)
    | UserDefined (pruneUDF, _) -> (pruneUDF box); None
    | PosSemDef p -> prunePosSemDef p box outer sample 
    | SubSet c -> pruneSubSet c box 
    | BoolIntv iv -> pruneBoolIntv iv box outer 
    | Atomic a -> pruneAtomic a box outer changed sample 
    | Conjunction p6 -> 
	if outer then pruneConjunction p6 box outer changed sample
        else pruneDisjunction p6 box outer changed sample; 
    | Disjunction l -> 
	if outer then pruneDisjunction l box outer changed sample
	else pruneConjunction l box outer changed sample; 
    | UnivQuant q -> 
	if outer then pruneUniv q box outer changed sample
        else pruneExist q box outer changed sample  
    | ExistQuant q -> 
	if outer then pruneExist q box outer changed sample 
	else pruneUniv q box outer changed sample 
    | Redundant q ->
	if outer then begin
	  ignore(prune_rec q.original_constr box outer changed sample);
	  ignore(prune_rec q.redundant_constr box outer changed sample);
	  None
	end else
	  prune_rec q.original_constr box outer changed sample;
    | UnivQuantBound q ->
	pruneUnivQuantBound q box outer changed sample
    | _ -> assert(false)
(*  assert(Constr.is_consistent constr box); *)
and prune (constr: constr) (box: box) (outer: bool) (changed: Intset.t) (sample: bool): witness = 
  let rounding_mode= get_rounding () in 
  let result= prune_rec constr box outer changed sample in
  set_rounding rounding_mode;  
  result

and prune_crude (constr: constr) (box: box) =
  let changed = Intset.empty_set () in 
  Intset.add changed (Util.int_list 0 ((dim box)-1));
  ignore (prune constr box true changed false);

and prune_crude_neg (constr: constr) (box: box) =
  let changed = Intset.empty_set () in 
  Intset.add changed (Util.int_list 0 ((dim box)-1));
  ignore (prune constr box false changed false)

and print_witness (w: witness): unit = 
  let print_psd_witness (w: float array) =
    let print_witness_el x = begin
      print_float x;
      print_string " "
    end in
    Array.iter print_witness_el w in
  match w with
    None -> () |
    Some (m, verified) -> 
      print_string (if verified then "Witness (not verified): "  else "Witness: ");
      print_psd_witness m;

  

(**************************************************************************************************************
	  the following currently is in prune.ml,
          but only because it refers back to pruning, 
          and O'Caml does not allowed circular dependencies between files
 **************************************************************************************************************)

and degree_truth (constr: constr) (point: float array): interval =
  (* assumption: each atomic constraint contains exactly one unary primitive constraint,
     let its variable be v
     eval computes the degree of truth of 'constr' at 'point'

     destroys auxvar information for 'constr'
   *)
  let atomic_info (constr: constr): variable * predicateUnary * constr =
    (* for an atomic constraint of the form t r 0, 
       returns the variable number corresponding to t,
       r, and
       the atomic constraint without the unary one (only the auxiliary constraints) *)

    let is_unary (c: constr): bool =
      match c with
	Unary _ -> true
      | _ -> false in

    (* atomic_info *)
    match constr with 
      Conjunction conn -> begin
	let unary = List.filter is_unary conn.branches in
	let repl = Conjunction { conn with branches = List.map (fun c -> if is_unary c then BoolIntv Bool_intv.true_intv else c) conn.branches } in
	assert((List.length unary)=1);
	match List.hd unary with
	  Unary u -> (u.var1, u.predicate1, repl)
	| _ -> assert(false) 
      end
    | _ -> assert(false)
    (* end atomic_info *) in
  
   (* degree_truth *) 
  let deg= 
    begin
      match constr with
	Atomic atomic -> begin
	  let constr= atomic.aconstr 
	  and box= cart_prod2 (Box.make (Array.map float_to_intv point)) (full_box (Box.dim atomic.obound)) in
	  let (v, p, degconstr)= atomic_info constr in
	  Constr.clear constr;
	  prune_crude degconstr box;
	  let term_val= get_bound box v in
	  match p with
	    Equal -> negII (absII term_val)
	  | LessEqual 
	  | Less -> negII term_val
	  | NotEqual -> absII term_val
	  | Greater 
	  | GreaterEqual -> term_val
	end
      | Conjunction conn ->
	  let degc = List.fold_left minIII { lo= infinity; hi= infinity } (List.map (function c -> degree_truth c point) conn.branches) in
	  degc
      | _ -> assert(false) 
    end in
(*  print_string "deg: "; Array.iter (fun x -> print_float x; print_string " ") point; Interval.print deg; print_newline (); *)
  deg

and compute_sample (constr: constr) (box: box) (neg: bool): witness =
(*  print_endline "---------------------sample";  *)
  let constr_copy = Constr.copy constr in
(*  Constr.clear constr_copy;*)
  let deg x = 
    let p = Interval.midpoint ((degree_truth constr_copy) x) in
    if neg then p else -. p in
  let sample= Local_search.search deg box in
  let check_sample = Box.init (dim box) (fun i -> float_to_intv sample.(i)) in
 (*  Constr.clear constr_copy; *)
(*   Box.print check_sample; print_endline "###";*)
  if !flag_sample_approximate then begin
    if neg then prune_crude_neg constr_copy check_sample
    else prune_crude constr_copy check_sample;
    if is_empty check_sample then None
    else Some (sample, false) (* unverified witness *)
  end else begin
    if neg then prune_crude constr_copy check_sample
    else prune_crude_neg constr_copy check_sample;
    if is_empty check_sample then Some (sample, true) (* verified sample *)
    else None
  end

(*
  let mid_point= midpoint box 
  and constr_copy = Constr.copy constr in
  Constr.clear constr_copy;
  if neg then prune_crude_neg constr_copy mid_point
  else prune_crude constr_copy mid_point;
  if is_empty mid_point then None
  else Some (Array.init (dim box) (function k -> Interval.midpoint (get_bound box k)))
*)
    
